import React from "react";

type StateArgs = {done?:boolean, open?:boolean};
type StateObj = {done:boolean, open:boolean};
type StateBinding = [state:StateObj, update:(args:StateArgs)=>void];

const CTX = React.createContext([{done:true, open:false}, (args)=>{}] as StateBinding);

export const Group =(props:{children:React.JSX.Element|React.JSX.Element[]})=>
{
    const [stateGet, stateSet] = React.useState({done:true, open:false} as StateObj);
    return <CTX.Provider value={[stateGet, (args:StateArgs)=> stateSet({...stateGet, ...args})]}>{props.children}</CTX.Provider>;
};

export const Menu =(props:{children:React.JSX.Element|React.JSX.Element[]})=>
{
    const [stateGet, stateSet] = React.useContext(CTX);
    const refElement:React.MutableRefObject<HTMLElement|null> = React.useRef( null );
    const refControl:React.MutableRefObject<CollapseControls|null> = React.useRef( null );

    type MenuClassStates = {Keep:string, Open:string, Shut:string, Move:string, Exit:string};
    const Classes:MenuClassStates =
    {
      Keep: "relative transition-all border",
      Open: "h-auto w-auto    top-10 duration-1000",
      Shut: "h-0    w-[300px] top-0  duration-1000",
      Move: "",
      Exit: "h-auto w-auto top-36 px-4 opacity-0"
    };
    const Window = window as {TwindInst?:(c:string)=>string};
    if(Window.TwindInst)
    {
      for(let stateName in Classes)
      {
        Classes[stateName as keyof MenuClassStates] = Window.TwindInst(Classes[stateName as keyof MenuClassStates]);
      }
    }

    React.useEffect(()=>refControl.current = Collapser(refElement.current as HTMLElement, true), []);
    React.useEffect(()=> {refControl.current && refControl.current(stateGet.open, 1000, ()=>stateSet({done:true}))}, [stateGet.open])

    return <div class={Classes.Keep} data-class-keep={Classes.Keep} data-class-open={Classes.Open} data-class-shut={Classes.Shut} ref={refElement as React.Ref<HTMLDivElement>}>
      { (!stateGet.open && stateGet.done) ? null : props.children}
    </div>;
};

export const Button =()=>
{
    const [stateGet, stateSet] = React.useContext(CTX);
    return <>
        <p>{JSON.stringify(stateGet)}</p>
        <button class="px-10 py-2 bg-red-500 text-white" onClick={e=>stateSet({open:true,  done:false})}>Open</button>
        <button class="px-10 py-2 bg-red-500 text-white" onClick={e=>stateSet({open:false, done:false})}>Close</button>
    </>;
};

type DoneCallback =(openState:boolean)=>void;
export type CollapseControls =(inOpen?:boolean, inMs?:number, inDone?:DoneCallback)=>void;
export function Collapser(inElement:HTMLElement, initialState = false)
{
    let userDone:DoneCallback = (openState) => {};
    let userMode = initialState;
    let frameRequest = 0;

    let transitions = new Set();

    const run = (inEvent:TransitionEvent)=>
    {
      if (inEvent.target == inElement)
      {
        console.log("transition", inEvent.target)
        transitions.add(inEvent.propertyName);
      }
    };

    const end = (inEvent:TransitionEvent)=>
    {
      if (inEvent.target == inElement)
      {
        transitions.delete(inEvent.propertyName);
        if(transitions.size === 0)
        {
          if (userMode)
          {
            inElement.setAttribute("style", "");
          }
          userDone(userMode);          
        }

      }
    };

    inElement.addEventListener("transitionend", end);
    inElement.addEventListener("transitionrun", run);

    return function(inOpen, inMs, inDone)
    {
      cancelAnimationFrame(frameRequest);

      if(arguments.length)
      {
        userDone = inDone|| ((m)=>{}) as DoneCallback;
        userMode = inOpen === true;

        // 1) capture current state and possible classes
        const h1 = inElement.clientHeight;
        const w1 = inElement.clientWidth;
        const classesInitial = inElement.getAttribute("class");
        const classesKeep = inElement.getAttribute("data-class-keep");
        const classesEnd = inElement.getAttribute(`data-class-${inOpen?"open":"shut"}`);
      
        // 2) disable transitions to get immediate results
        inElement.setAttribute("style", "");
        inElement.style.transition = "width none, height none";

        // 3) add new classes and capture new height
        inElement.className = classesKeep + " " + classesEnd;
        const h2 = inElement.clientHeight;
        const w2 = inElement.clientWidth;

        // 4) put everything back (and turn on overflow hidden)
        inElement.style.height = h1+"px";
        inElement.style.width = w1+"px";
        inElement.style.overflow = "hidden";
        inElement.className = classesInitial ?? "";

        console.log(`BEFORE classes:"${inElement.className}" width:"${w1}" height:"${h1}"`);

        frameRequest = requestAnimationFrame(()=>
        {
          inElement.style.transition = "";

          frameRequest = requestAnimationFrame(()=>
          {
            inElement.style.height = `${h2}px`;
            inElement.style.width = `${w2}px`;
            inElement.className = classesKeep + " " + classesEnd;
            console.log(`AFTER classes:"${inElement.className}" width:"${w2}" height:"${h2}"`);
          });
          
        });
      }
      else
      {
        inElement.removeEventListener("transitionrun", run);
        inElement.removeEventListener("transitionend", end);
      }
    } as CollapseControls;
}